// file: hmm_cmdl.cc
//
// command line processing methods
//

// system include files
//
#include <stdlib.h>

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"
#include "../include/hmm_constants.h"

//-----------------------------------------------------------------------------
// method: get_parameters_cc
//
// arguments:
//
//  int argc: argc from the main program
//  char** argv: argv from the main program
//
//  int_4& S: (output) number of coins used in the model
//  int_4& K: (output) number of output symbols in the model
//  int_4& P: (output) number of training passes
//  int_4& L: (output) number of sequences to generate
//  int_4& mode: (output) processing mode (test/train/adapt)
//  int_4& alg: (output) algorithm (viterbi/baum welch/...)
//  int_4& debug: (output) debug level
//
//  char_1* file1: (output) observation sequences
//  char_1* file2: (output) hmm model
//
// return value: a logical_1 indicating status
//
// this function parses command line arguments and returns appropriate values
//
//-----------------------------------------------------------------------------
logical_1 Hmm::get_parameters_cc(int argc, char** argv,
				 int_4& S, int_4& K, int_4& P, int_4& L,
				 int_4& mode, int_4& alg, int_4& debug,
				 char_1* file1, char_1* file2) {

  // perform some basic checks - there must be two arguments plus the
  // program name
  //
  if (argc < 3)
    {help_cc();}

  // pick off the filenames
  //
  strcpy((char*)file1, argv[argc-2]);
  strcpy((char*)file2, argv[argc-1]);

  // main decoding loop
  //
  int_4 last_index = argc - (int_4)2;

  for (int_4 index=1; index<last_index; index++) {

    // pick off the number of states
    //
    if (strcmp(ISIP_OPT_S, argv[index]) == (int)0) {
      S = atoi(argv[index+(int)1]);
      if (S < (int_4)0)
	{return ISIP_FALSE;}
      index++;
    }

    // pick off the number of output symbols
    //
    else if (strcmp(ISIP_OPT_K, argv[index]) == (int)0) {
      K = atoi(argv[index+(int)1]);
      if (K < (int_4)0)
	{return ISIP_FALSE;}
      index++;
    }

    // pick off the number of iterations
    //
    else if (strcmp(ISIP_OPT_P, argv[index]) == (int)0) {
      P = atoi(argv[index+(int)1]);
      if (P < (int_4)0)
	{return ISIP_FALSE;}
      index++;
    }

    // pick off the number of sequences
    //
    else if (strcmp(ISIP_OPT_L, argv[index]) == (int)0) {
      L = atoi(argv[index+(int)1]);
      if (L < (int_4)0)
	{return ISIP_FALSE;}
      index++;
    }

    // pick off the algorithm
    //
    else if (strcmp(ISIP_OPT_ALG_1, argv[index]) == (int)0) {
      alg = ISIP_ALG_VITERBI;
    }
    else if (strcmp(ISIP_OPT_ALG_2, argv[index]) == (int)0) {
      alg = ISIP_ALG_BAUMWELCH;
    }

    // pick off mode
    //
    else if (strcmp(ISIP_OPT_MODE_1, argv[index]) == (int)0) {
      mode = ISIP_MODE_TRAIN;
    }
    else if (strcmp(ISIP_OPT_MODE_2, argv[index]) == (int)0) {
      mode = ISIP_MODE_TEST;
    }
    else if (strcmp(ISIP_OPT_MODE_3, argv[index]) == (int)0) {
      mode = ISIP_MODE_GENERATE;
    }
    else if (strcmp(ISIP_OPT_MODE_4, argv[index]) == (int)0) {
      mode = ISIP_MODE_UPDATE;
    }

    // pick off the debug level
    //
    else if (strcmp(ISIP_OPT_DBG, argv[index]) == (int)0) {
      debug = atoi(argv[index+(int)1]);
      if ((debug < ISIP_DEBUG_NONE) || (debug > ISIP_DEBUG_FULL))
	{return ISIP_FALSE;}
      index++;
    }
    
    // check for a help flag
    //
    else if (strcmp(ISIP_OPT_HELP, argv[index]) == (int)0) {
      
      // execute the help command
      //
      help_cc();
    }

    // else: unknown argument
    //
    else
      {return ISIP_FALSE;}
  }

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: help_cc
//
// arguments: none
//
// return value: none
//
// a real cheesy way to do a help message - but fairly transportable
//
//-----------------------------------------------------------------------------
volatile void Hmm::help_cc() {

  // display a hardcoded help message
  //
  fprintf(stdout, "\n");
  fprintf(stdout, "program: hmm.cc\n");
  fprintf(stdout, "synopsis: hmm.exe [options] file1 file2\n");
  fprintf(stdout, "descr: trains and evaluates a discrete hmm\n");
  fprintf(stdout, "example: hmm.exe -test data.text model.text\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "options:\n");
  fprintf(stdout, "\n");
  fprintf(stdout, " options requiring numeric values:\n");
  fprintf(stdout, " \n");
  fprintf(stdout, "  S: (input) number of states to use in the model (default=2)\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, "  K: (input) number of symbols to use in the model (default=2)\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, "  P: (input) number of passed on the training data (default=10)\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, "  debug: (input) output debugging information (0=none, 3=full)\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, " options functioning as flags (no numeric values required):\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, "  viterbi: (input) use the viterbi algorithm for probability computations\n");
  fprintf(stdout, "  baumwelch: (input) use the baum-welch algorithm\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, "  train: (input) train a model and output it to file2\n");
  fprintf(stdout, "  test: (input) read a model from file2, evaluate the data in file1\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, "arguments:\n");
  fprintf(stdout, "\n");
  fprintf(stdout, " file1: (input) file containing observation sequences\n");
  fprintf(stdout, " file2: (input/output) file containing an hmm model\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "this program implements a simple discrete hmm. the program is intended to\n");
  fprintf(stdout, "support basic experiments on discrete hmms as described in numerous\n");
  fprintf(stdout, "textbooks. for example, you can simulate coin-toss experiments and\n");
  fprintf(stdout, "experiment with models using different numbers of coins.\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "two somewhat important details of the program are:\n");
  fprintf(stdout, "\n");
  fprintf(stdout, " - a fully ergodic hmm is assumed (all states are fully connected to\n");
  fprintf(stdout, "   all other states\n");
  fprintf(stdout, " \n");
  fprintf(stdout, " - initial values are randomly assigned (this causes fits for the viterbi\n");
  fprintf(stdout, "   algorithm\n");
  fprintf(stdout, "\n");
  fprintf(stdout, "An example of how to run the program is given below:\n");
  fprintf(stdout, " \n");
  fprintf(stdout, " training mode:\n");
  fprintf(stdout, " \n");
  fprintf(stdout, "  hmm.exe -S 2 -K 2 -train -baumwelch data.text model.text\n");
  fprintf(stdout, " \n");
  fprintf(stdout, "  this requests a 2 state hmm using 2 output symbols per state to be trained\n");
  fprintf(stdout, "  using the baum-welch algorithm. the resulting model will be written\n");
  fprintf(stdout, "  to model.text\n");
  fprintf(stdout, "  \n");
  fprintf(stdout, " testing mode:\n");
  fprintf(stdout, " \n");
  fprintf(stdout, "  hmm.exe -test -baumwelch data.text model.text\n");
  fprintf(stdout, " \n");
  fprintf(stdout, " note that it is possible to do interesting things such as train using the\n");
  fprintf(stdout, " viterbi algorithm and test using the baumwelch algorithm, or evaluate\n");
  fprintf(stdout, " a model on new data, and compare the result to data used to train the\n");
  fprintf(stdout, " model (similar to the speaker independent speech recognition problem).\n");
  fprintf(stdout, " \n");
  fprintf(stdout, "enjoy!  \n");
  fprintf(stdout, "\n");

  // exit ungracefully
  //
  exit(ISIP_NO_ERROR);
}
