// file: hmm_disp.cc
//
// methods that display various aspects of the model
//

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"

//-----------------------------------------------------------------------------
// method: display_cc
//
// arguments:
//
//  FILE* fp: (input) an open file pointer (typically stdout)
//
// return value: a logical_1 indicating status
//
// this function displays the information contained
// in a model.
//
//-----------------------------------------------------------------------------
logical_1 Hmm::display_cc(FILE* fp) {

  // process the dimension information
  //
  fprintf(fp, "\nhidden markov model:\n");
  fprintf(fp, "  S (number of states) = %ld\n", S);
  fprintf(fp, "  K (number of output symbols) = %ld\n", K);

  // process the symbol table
  //
  if (display_symbol_table_cc(fp) == ISIP_FALSE)
    {return ISIP_FALSE;}

  // process the transition probs
  //
  fprintf(fp, "a[%ldx%ld] (transition probabilities):\n", S, S);
  for (int_4 i=0; i<S; i++) {
    for (int_4 j=0; j<S; j++) {
      fprintf(fp, "\ta[%ld][%ld]: state %ld --> state %ld = %f\n",
	      i, j, i, j, a[i][j]);
    }
  }

  // process the output probs
  //
  fprintf(fp, "b[%ldx%ld] (output probabilities):\n", S, K);
  for (int_4 i=0; i<S; i++) {
    for (int_4 j=0; j<K; j++) {
      fprintf(fp, "\tb[%ld][%ld]: P(%1c) = %f\n",
	      i, j, y[j], b[i][j]);
    }
  }

  // process the initial state probs
  //
  fprintf(fp, "pi[%ld] (initial state probabilities):\n", S);
  for (int_4 i=0; i<S; i++) {
    fprintf(fp, "\tpi[%ld]: state %ld = %f\n", i, i, pi[i]);
  }
  fprintf(fp, "\n");  

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: display_symbol_table_cc
//
// arguments:
//
//  FILE* fp: (input) an open file pointer (typically stdout)
//
// return value: a logical_1 indicating status
//
// this function displays the information contained in the symbol table
//
//-----------------------------------------------------------------------------
logical_1 Hmm::display_symbol_table_cc(FILE* fp) {

  // process the symbol table
  //
  fprintf(fp, "y[%ld] (symbol table):\n", K);
  for (int_4 i=0; i<K; i++) {
    fprintf(fp, "\ty[%ld] %1c\n", i, y[i]);
  }

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: display_sequence_info_cc
//
// arguments:
//
//  int_4 seq_num: (input) obs sequence identifier
//  float_4 seq_prob: (input) the log prob of the sequence
//  int_4 T: (input) sequence length
//  char_1* O_sym: (input) observation sequence in symbols
//  int_4* O: (input) observation sequence in indices
//  int_4* x: (input) optimal state sequence
//  int_4 debug: (input) debug flag
//  FILE* fp: (input) an open file pointer (typically stdout)
//
// return value: a logical_1 indicating status
//
// this function displays the information about a sequence after it has
// been processed
//
//-----------------------------------------------------------------------------
logical_1 Hmm::display_sequence_info_cc(int_4 seq_num, float_4 seq_prob,
					int_4 T, char_1* O_sym, int_4* O,
					int_4* x, int_4 debug, FILE* fp) {

  // display a summary of the prob information
  //
  fprintf(fp, "\tlog P(O[%ld]/M) (%f): %s\n", seq_num, seq_prob, O_sym);

  // display the converted utterance
  //
  if (debug > ISIP_DEBUG_BRIEF) {

    // echo the original input
    //
    fprintf(fp, "\torig input: ");
    for (int_4 i=0; i<T; i++)
      {fprintf(fp, "  %1c", O_sym[i]);}

    // echo the symbol table indices
    //
    fprintf(fp, "\n\tconv input: ");
    for (int_4 i=0; i<T; i++)
      {fprintf(fp, "%3ld", O[i]);}
    fprintf(fp, "\n");
  }

  // echo the state sequence
  //
  if (debug > ISIP_DEBUG_NONE) {
    fprintf(fp, "\tstate sequence = ");
    for (int_4 i=0; i<T; i++) {
      fprintf(fp, "%ld ", x[i]);
    }
    fprintf(fp, "\n");
  }

  // exit gracefully
  //
  return ISIP_TRUE;
}
