// file: hmm_gen.cc
//

// system include files
//
#include <stdlib.h>

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"
#include "../include/hmm_constants.h"

//-----------------------------------------------------------------------------
// method: generate_cc
//
// arguments:
//  int_4& T: (input) sequence length
//  char_1* O_sym: (input) observation sequence in symbols
//  int_4* O: (input) observation sequence in indices
//  int_4* x: (input) optimal state sequence
//
// return: a float_4 value containing P(Y|M)
//
// this method traverses a model and generates random output that
// follows the statistical model embedded in the model.
//
//-----------------------------------------------------------------------------
float_4 Hmm::generate_cc(int_4& T, char_1* O_sym, int_4* O, int_4* x) {

  // randomly select a duration - use a simple uniform duration rule
  //
  int_4 max_dur = (int_4)pow(S,S);
  float_4 prob_dur[max_dur];
  for (int_4 i=0; i<max_dur; i++) {
    prob_dur[i] = 1.0 / (float_4)max_dur;
  }
  T = generate_node_cc(prob_dur, max_dur) + (int_4)1;

  // loop until done
  //
  float_4 prob = 0.0;

  for (int_4 t=0; t<T; t++) {

    // advance one node
    //
    if (t == 0) {
      x[t] = generate_node_cc(pi, S);
      prob += log10_cc(pi[x[t]]);
    }
    else {
      x[t] = generate_node_cc(a[x[t-1]], S);
      prob += log10_cc(a[x[t-1]][x[t]]);      
    }

    // generate an output symbol
    //
    O[t] = generate_node_cc(b[x[t]], K);
    O_sym[t] = y[O[t]];
    prob += log10_cc(b[x[t]][O[t]]);
  }

  // exit gracefully
  //
  O_sym[T] = ISIP_NULL;
  return prob;
}

//-----------------------------------------------------------------------------
// method: generate_node_cc
//
// arguments:
//  float_4* v: (input) probability distribution
//  int_4 nv: (input) dimension of v
//
// return: an int_4 value containing an index
//
// this method attempts to generate random integers in proportion to
// a user-specified probability distribution. it is based on the
// unix drand48 function (not the most portable these days).
//
//-----------------------------------------------------------------------------
int_4 Hmm::generate_node_cc(float_4* v, int_4 nv) {

  // generate a random number between [0,1]
  //
  float_4 x;
  x = drand48();

  // initialize
  //
  float_4 sum = v[0];
  if (x < sum)
    {return (int_4)0;}

  // loop over all elements in the vector
  //
  for (int_4 i=1; i<(nv-1); i++) {
    sum += v[i];
    if (x < sum)
      {return i;}
  }

  // exit gracefully
  //
  return (nv-1);
}
