// file: hmm_io.cc
//
// methods involving various types of i/o

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"

//-----------------------------------------------------------------------------
// method: load_cc
//
// arguments:
//
//  char_1* file: (input) file containing a trained hmm
//
// return value: a logical_1 indicating status
//
// this function loads a model from a file
//
//-----------------------------------------------------------------------------
logical_1 Hmm::load_cc(char_1* file) {

  // open the file
  //
  FILE* fp = fopen((char*)file, "r");
  if (fp == (FILE*)NULL)
    {return ISIP_FALSE;}

  // read general parameters
  //
  fscanf(fp, "%*s");
  fscanf(fp, "%*s%*s%ld", &S);
  fscanf(fp, "%*s%*s%ld", &K);

  // allocate space
  //
  if (allocate_cc(S, K) == ISIP_FALSE)
    {return ISIP_FALSE;}

  // read the symbol table
  //
  fscanf(fp, "%*s");

  for (int_4 i=0; i<K; i++) {
    fscanf(fp, "%*d%1s", &y[i]);
  }

  // read the transition probs
  //
  fscanf(fp, "%*s");

  for (int_4 i=0; i<S; i++) {
    for (int_4 j=0; j<S; j++) {
      fscanf(fp, "%*d%*d%f", &(a[i][j]));
    }
  }

  // read the output symbol probs
  //
  fscanf(fp, "%*s");

  for (int_4 i=0; i<S; i++) {
    for (int_4 j=0; j<K; j++) {
      fscanf(fp, "%*d%*d%f", &(b[i][j]));
    }
  }

  // read the initial state probs
  //
  fscanf(fp, "%*s");
  
  for (int_4 i=0; i<S; i++) {
    fscanf(fp, "%*d%f", &(pi[i]));
  }

  // close the file
  //
  fclose(fp);

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: convert_cc
//
// arguments:
//  int_4* O: (output) buffer containing the input observation sequence
//                       represented as indices in the symbol table
//  int_4& T: (output) "true" length of the input sequence
//  char_1* O_sym: (output) buffer containing the condensed input
//				 sequence with all garbage characters removed
//  char_1* line_buffer: (input) the input data
//
// return: a logical_1 value indicating the status
//
// this method converts an input sequence, such as "HTHT  HTHT"
// to a condensed sequence "HTHTHTHT" and "01010101", where 0 is represented
// by "H" in the symbol table.
//
// note that O and O_sym must be allocated prior to calling this method.
// also note that the conversion ignores all characters not in the symbol
// table.
//-----------------------------------------------------------------------------
logical_1 Hmm::convert_cc(int_4* O, int_4& T,
			  char_1* O_sym, char_1* line_buffer) {

  // loop over all elements
  //
  int_4 length = strlen((char*)line_buffer);
  T = 0;

  for (int_4 i=0; i<length; i++) {

    // check if the symbol is on the list
    //
    int_4 pos;

    if ((pos = match_cc(line_buffer[i])) >= (int_4)0) {
      
      // add it to the output string
      //
      O[T] = pos;
      O_sym[T] = line_buffer[i];
      T++;

    }
  }

  // null-terminate all strings
  //
  int_4 last_char = strlen((char*)line_buffer);
  line_buffer[last_char-1] = ISIP_NULL;
  O_sym[T] = ISIP_NULL;

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: save_cc
//
// arguments:
//
//  char_1* file: (input) filename to save model to
//
// return value: a logical_1 indicating status
//
// this function writes an hmm to a file
//
//-----------------------------------------------------------------------------
logical_1 Hmm::save_cc(char_1* file) {

  // open the file
  //
  FILE* fp = fopen((char*)file, "w");
  if (fp == (FILE*)NULL)
    {return ISIP_FALSE;}

  // process the dimension information
  //
  fprintf(fp, "M:\n");
  fprintf(fp, "  S = %ld\n", S);
  fprintf(fp, "  K = %ld\n", K);

  // process the symbol table
  //
  fprintf(fp, "y[%ld]:\n", K);
  for (int_4 i=0; i<K; i++) {
    fprintf(fp, "\t%ld %1c\n", i, y[i]);
  }

  // process the transition probs
  //
  fprintf(fp, "a[%ldx%ld]:\n", S, S);
  for (int_4 i=0; i<S; i++) {
    for (int_4 j=0; j<S; j++)
      {fprintf(fp, "\t%ld %ld %f\n", i, j, a[i][j]);}
  }

  // process the output probs
  //
  fprintf(fp, "b[%ldx%ld]:\n", S, K);
  for (int_4 i=0; i<S; i++) {
    for (int_4 j=0; j<K; j++)
      {fprintf(fp, "\t%ld %ld %f\n", i, j, b[i][j]);}
  }

  // process the initial state probs
  //
  fprintf(fp, "pi[%ld]:\n", S);
  for (int_4 i=0; i<S; i++) {
    fprintf(fp, "\t%ld %f\n", i, pi[i]);
  }

  // close the file
  //
  fclose(fp);

  // exit gracefully
  //
  return ISIP_TRUE;
}
