// file: hmm.cc
//

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"
#include "../include/hmm_constants.h"

//-----------------------------------------------------------------------------
//
// program: hmm.cc
// synopsis: hmm.exe [options] file1 file2
// descr: trains and evaluates a discrete hmm
// example: hmm.exe -test data.text model.text
//
// options:
//
//  options requiring numeric values:
//  
//   S: (input) number of states to use in the model (default=2)
//
//   K: (input) number of symbols to use in the model (default=2)
//
//   P: (input) number of passed on the training data (default=10)
//
//   L: (input) number of observation sequences (default=100)
//
//   debug: (input) output debugging information (0=none, 3=full)
//
//  options functioning as flags (no numeric values required):
//
//   viterbi: (input) use the viterbi algorithm for probability computations
//   baumwelch: (input) use the baum-welch algorithm
//
//   train: (input) train a model and output it to file2
//   test: (input) read a model from file2, evaluate the data in file1
//   generate: (input) generate data from a model and output it to file1
//   update: (input) update an existing model
//
// arguments:
//
//  file1: (input) file containing observation sequences
//  file2: (input/output) file containing an hmm model
//
// this program implements a simple discrete hmm. the program is intended to
// support basic experiments on discrete hmms as described in numerous
// textbooks. for example, you can simulate coin-toss experiments and
// experiment with models using different numbers of coins.
//
// two somewhat important details of the program are:
//
//  - a fully ergodic hmm is assumed (all states are fully connected to
//    all other states
//
//  - initial values are randomly assigned (this causes fits for the viterbi
//    algorithm
//
// An example of how to run the program is given below:
//
//  training mode:
//
//   hmm.exe -S 2 -K 2 -train -baumwelch data.text model.text
//
//   this requests a 2 state hmm using 2 output symbols per state to be trained
//   using the baum-welch algorithm. the resulting model will be written
//   to model.text
//
//  testing mode:
//
//   hmm.exe -test -baumwelch data.text model.text
//
// note that it is possible to do interesting things such as train using the
// viterbi algorithm and test using the baumwelch algorithm, or evaluate
// a model on new data, and compare the result to data used to train the
// model (similar to the speaker independent speech recognition problem).
//
// enjoy!  
//
//-----------------------------------------------------------------------------
main(int argc, char **argv) {

  //---------------------------------------------------------------------------
  //
  // variable declarations
  //
  //---------------------------------------------------------------------------

  // declare options
  //
  int_4 S = ISIP_DEFAULT_S;
  int_4 K = ISIP_DEFAULT_K;
  int_4 P = ISIP_DEFAULT_P;
  int_4 L = ISIP_DEFAULT_L;
  int_4 alg = ISIP_DEFAULT_ALG;
  int_4 mode = ISIP_DEFAULT_MODE;
  int_4 debug = ISIP_DEFAULT_DBG;

  char_1 file1[ISIP_MAX_FNAME_SIZE];
  char_1 file2[ISIP_MAX_FNAME_SIZE];

  // declare an hmm object
  //
  Hmm hmm;

  // declare some scratch space for the observation vectors
  //
  FILE* fp;
  char_1 line_buffer[ISIP_MAX_LINE_SIZE];
  char_1 O_sym[ISIP_MAX_LINE_SIZE];
  int_4 T;
  int_4 O[ISIP_MAX_LINE_SIZE];

  // declare some space to hold recognition output
  //
  int_4 seq_num = 0;
  int_4 x[ISIP_MAX_LINE_SIZE];
  float_4 seq_prob;
  float_4 total_prob = 0.0;

  //---------------------------------------------------------------------------
  //
  // initialization
  //
  //---------------------------------------------------------------------------

  if (hmm.get_parameters_cc(argc, argv,
			    S, K, P, L, mode, alg, debug,
			    file1, file2) == ISIP_FALSE) {
    fprintf(stdout, "%s: error decoding command line\n", ISIP_PROGRAM_NAME);
    return(ISIP_ERROR);
  }

  // load a model if necessary
  //
  if ((mode == ISIP_MODE_TEST) || (mode == ISIP_MODE_UPDATE)) {
    if (hmm.load_cc(file2) == ISIP_FALSE) {
      fprintf(stdout, "%s: error loading data from file %s\n",
	      ISIP_PROGRAM_NAME, file2);
      return(ISIP_ERROR);
    }

    if (debug > ISIP_DEBUG_NONE)
      {hmm.display_cc(stdout);}
  }

  // open the data file
  //
  if (mode != ISIP_MODE_GENERATE) {
    if ((fp = fopen((char*)file1, "r")) == ISIP_FALSE) {
      fprintf(stdout, "%s: error opening data file %s\n",
	      ISIP_PROGRAM_NAME, file1);
      return(ISIP_ERROR);
    }
  }
  else {
    if ((fp = fopen((char*)file1, "w")) == ISIP_FALSE) {
      fprintf(stdout, "%s: error opening data file %s\n",
	      ISIP_PROGRAM_NAME, file1);
      return(ISIP_ERROR);
    }
  }

  //---------------------------------------------------------------------------
  //
  // generation
  //
  //---------------------------------------------------------------------------

  if (mode == ISIP_MODE_GENERATE) {

    // load a model
    //
    if (hmm.load_cc(file2) == ISIP_FALSE) {
      fprintf(stdout, "%s: error loading data from file %s\n",
	      ISIP_PROGRAM_NAME, file2);
      return(ISIP_ERROR);
    }

    total_prob = 0.0;
    for (seq_num=0; seq_num<L; seq_num++) {
      
      // generate a sequence
      //
      float_4 seq_prob = hmm.generate_cc(T, O_sym, O, x);
      total_prob += seq_prob;

      // provide some debugging information to stdout
      //
      hmm.display_sequence_info_cc(seq_num, seq_prob,
				   T, O_sym, O, x, debug, stdout);

      // write to the file
      //
      fprintf(fp, "%s\n", O_sym);
    }

    // summarize the output
    //
    fprintf(stdout, "\nlog P(O/M) = %f (%ld sequences)\n",
	    total_prob, seq_num);
  }

  //---------------------------------------------------------------------------
  //
  // training or update
  //
  //---------------------------------------------------------------------------

  if ((mode == ISIP_MODE_TRAIN) || (mode == ISIP_MODE_UPDATE)) {

    // initialize the model - only if the training mode has been requested
    //
    if (mode == ISIP_MODE_TRAIN) {
      if (hmm.init_cc(S, K) == ISIP_FALSE) {
	fprintf(stdout, "%s: error initializing model\n", ISIP_PROGRAM_NAME);
	return(ISIP_ERROR);
      }      

      // seed the model
      //
      if (hmm.seed_cc() == ISIP_FALSE) {
	fprintf(stdout, "%s: error initializing model\n", ISIP_PROGRAM_NAME);
	return(ISIP_ERROR);
      }      

      if (debug > ISIP_DEBUG_DETAILED) {
	fprintf(stdout, "initalized model:\n");
	hmm.display_cc(stdout);
      }

      // generate the symbol table by processing all of the input data
      //
      if (hmm.train_symbol_table_cc(fp) == ISIP_FALSE) {
	fprintf(stdout, "%s: error training symbol table\n",
		ISIP_PROGRAM_NAME);
	return(ISIP_ERROR);
      }      

      if (debug > ISIP_DEBUG_BRIEF) {
	hmm.display_symbol_table_cc(stdout);
      }
    }
      
    // perform P passes on the data
    //
    for (int_4 pass=0; pass<P; pass++) {

      fprintf(stdout, "\npass no. %ld:\n", pass);
      total_prob = 0;

      // initialize counters
      //
      if (hmm.init_counters_cc() == ISIP_FALSE) {
	fprintf(stdout, "%s: error initializing counters\n",
		ISIP_PROGRAM_NAME);
	return(ISIP_ERROR);
      }      

      // cycle through all of the data in the file
      //
      seq_num = 0;
      rewind(fp);
    
      while (fgets((char*)line_buffer,ISIP_MAX_LINE_SIZE,fp) != (char*)NULL) {

	// transform the data to symbol indices
	//
	if (hmm.convert_cc(O, T, O_sym, line_buffer) == ISIP_FALSE) {
	  fprintf(stdout, "%s: error converting data item [%s]\n",
		  ISIP_PROGRAM_NAME, line_buffer);
	}

	// process non-empty strings
	//
	if (T > (int_4)0) {

	  // perform an update pass
	  //
	  seq_prob = hmm.train_cc(x, O, T, alg, debug);
	  total_prob += seq_prob;

	  // provide some debugging information
	  seq_num++;
	  hmm.display_sequence_info_cc(seq_num, seq_prob,
				       T, O_sym, O, x, debug, stdout);
	}
      }

      // update the model
      //
      if (hmm.replace_cc(alg) == ISIP_FALSE) {
	fprintf(stdout, "%s: error replacing model\n", ISIP_PROGRAM_NAME);
	return(ISIP_ERROR);
      }

      // display the current model
      //
      if (debug > ISIP_DEBUG_NONE) {
	fprintf(stdout, "\ncurrent model:\n");
	hmm.display_cc(stdout);
      }

      // summarize the output
      //
      fprintf(stdout, "\nlog P(O/M) = %f (%ld sequences)\n",
	      total_prob, seq_num);

      // save model after each iteration
      //
      if (hmm.save_cc(file2) == ISIP_FALSE) {
	fprintf(stdout, "%s: error writing model\n", ISIP_PROGRAM_NAME);
	return(ISIP_ERROR);
      }
    }
  }

  //---------------------------------------------------------------------------
  //
  // testing
  //
  //---------------------------------------------------------------------------

  if (mode == ISIP_MODE_TEST) {

    // cycle through all of the data
    //
    seq_num = 0;
    total_prob = 0;
    fprintf(stdout, "processing input data:\n");
    rewind(fp);

    while (fgets((char*)line_buffer, ISIP_MAX_LINE_SIZE, fp) != (char*)NULL) {

      // transform the data to symbol indices
      //
      if (hmm.convert_cc(O, T, O_sym, line_buffer) == ISIP_FALSE) {
	fprintf(stdout, "%s: error converting data item [%s]\n",
		ISIP_PROGRAM_NAME, line_buffer);
      }

      // process non-empty strings
      //
      if (T > (int_4)0) {
	fprintf(stdout, "\n");

	// recognize the sequence
	//
	seq_prob = hmm.test_cc(x, O, T, alg, debug);
	total_prob += seq_prob;

	// provide some debugging information
	seq_num++;
	hmm.display_sequence_info_cc(seq_num, seq_prob,
				     T, O_sym, O, x, debug, stdout);
      }
    }

    // summarize the output
    //
    fprintf(stdout, "\nlog P(O/M) = %f (%ld sequences)\n",
	    total_prob, seq_num);
  }

  //---------------------------------------------------------------------------
  //
  // exit
  //
  //---------------------------------------------------------------------------

  // close the data file
  //
  fclose(fp);

  // exit gracefully
  //
  return(ISIP_NO_ERROR);
}
