// file: hmm_mem.cc
//
// methods that manage model memory
//

// system include files
//
#include <memory.h>

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"

//-----------------------------------------------------------------------------
// method: cleanup_cc
//
// arguments: none
//
// return: none
//
// this method deletes the space used by an old model, preparing
// space for a new model
//
//-----------------------------------------------------------------------------
logical_1 Hmm::cleanup_cc() {

  // delete the 2d arrays associated with the model
  //
  if (S > (int_4)0) {

    // delete transition probs
    //
    if (a != (float_4**)NULL) {
      for (int_4 i=0; i<S; i++) {
	if (a[i] != (float_4*)NULL)
	  {delete [] a[i];}
      }
    }

    // delete output probs
    //
    if (b != (float_4**)NULL) {
      for (int_4 i=0; i<S; i++) {
	if (b[i] != (float_4*)NULL)
	  {delete [] b[i];}
      }
    }
  }

  // delete the 1D arrays associated with the model
  //
  if (y != (char_1*)NULL)
    {delete [] y;}
  if (a != (float_4**)NULL)
    {delete [] a;}
  if (b != (float_4**)NULL)
    {delete [] b;}
  if (pi != (float_4*)NULL)
    {delete [] pi;}

  // set important parameters to zero
  //
  S = (int_4)0;
  K = (int_4)0;

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: cleanup_counters_cc
//
// arguments: none
//
// return: none
//
// this method deletes the space used to update a model
//
//-----------------------------------------------------------------------------
logical_1 Hmm::cleanup_counters_cc() {

  // delete the 2d arrays associated with the model
  //
  if (S > (int_4)0) {

    // delete transition probs
    //
    if (a_new != (float_4**)NULL) {
      for (int_4 i=0; i<S; i++) {
	if (a_new[i] != (float_4*)NULL)
	  {delete [] a_new[i];}
      }
    }

    // delete output probs
    //
    if (b_new != (float_4**)NULL) {
      for (int_4 i=0; i<S; i++) {
	if (b_new[i] != (float_4*)NULL)
	  {delete [] b_new[i];}
      }
    }
  }

  // delete the 1D arrays associated with the model
  //
  if (a_new != (float_4**)NULL)
    {delete [] a_new;}
  if (b_new != (float_4**)NULL)
    {delete [] b_new;}
  if (pi_new != (float_4*)NULL)
    {delete [] pi_new;}

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: allocate_cc
//
// arguments:
//
//  int_4 new_S: (input) number of coins used in the model
//  int_4 new_K: (input) number of output symbols in the model
//
// return value: a logical_1 indicating status
//
// this function allocates internal space for a model
//
//-----------------------------------------------------------------------------
logical_1 Hmm::allocate_cc(int_4 new_S, int_4 new_K) {

  // save dimensions
  //
  S = new_S;
  K = new_K;

  // create space for the symbol table
  //
  y = new char_1[K];
  memset(y, (int)0, K*sizeof(char_1));

  // create space for the transition probs
  //
  a = new float_4*[S];
  for (int_4 i=0; i<S; i++) {
    a[i] = new float_4[S];
    memset(a[i], (int)0, S*sizeof(float_4));
  }

  // create space for the obs probs
  //
  b = new float_4*[S];
  for (int_4 i=0; i<S; i++) {
    b[i] = new float_4[K];
    memset(b[i], (int)0, K*sizeof(float_4));
  }

  // create space for the initial state probs
  //
  pi = new float_4[S];
  memset(pi, (int)0, S*sizeof(float_4));

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: allocate_counters_cc
//
// arguments: none
//
// return value: a logical_1 indicating status
//
// this function allocates internal space for updating a model
//
//-----------------------------------------------------------------------------
logical_1 Hmm::allocate_counters_cc() {

  // create space for the transition probs
  //
  a_new = new float_4*[S];

  for (int_4 i=0; i<S; i++) {
    a_new[i] = new float_4[S];
    memset(a_new[i], (int)0, S*sizeof(float_4));
  }

  // create space for the obs probs
  //
  b_new = new float_4*[S];

  for (int_4 i=0; i<S; i++) {
    b_new[i] = new float_4[K];
    memset(b_new[i], (int)0, K*sizeof(float_4));
  }

  // create space for the initial state probs
  //
  pi_new = new float_4[S];
  memset(pi_new, (int)0, S*sizeof(float_4));

  // exit gracefully
  //
  return ISIP_TRUE;
}
