// file: hmm_test.cc
//

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"
#include "../include/hmm_constants.h"

//-----------------------------------------------------------------------------
// method: test_cc
//
// arguments:
//  int_4* x: (output) optimal state sequence
//  int_4* O: (input) observation sequence
//  int_4 T: (input) length of the input sequence
//  int_4 alg: (input) algorithm flag
//  int_4 debug: (input) debug flag
//
// return: a float_4 value containing the log prob of the input sequence
//
//-----------------------------------------------------------------------------
float_4 Hmm::test_cc(int_4* x, int_4* O, int_4 T, int_4 alg, int_4 debug) {

  // branch on the algorithm
  //
  if (alg == ISIP_ALG_VITERBI) {
    return test_viterbi_cc(x, O, T, debug);
  }
  else {
    return test_baumwelch_cc(x, O, T, debug);
  }
}

//-----------------------------------------------------------------------------
// method: test_viterbi_cc
//
// arguments:
//  int_4* x: (output) optimal state sequence
//  int_4* O: (input) observation sequence
//  int_4 T: (input) length of the input sequence
//  int_4 debug: (input) debug flag
//
// return: a float_4 value containing the log prob of the input sequence
//
//-----------------------------------------------------------------------------
float_4 Hmm::test_viterbi_cc(int_4* x, int_4* O, int_4 T, int_4 debug) {

  // the dp solution requires an array with a dimension of T x S
  //
  float_4 s_err[T][S];
  int_4 bkptr[T][S];

  // initialize the costs - note that the row dimension of s_err corresponds
  // to a state, and the column dimension corresponds to time.
  //
  for (int_4 i=0; i<S; i++) {
    s_err[0][i] = log10_cc(pi[i]) + log10_cc(b[i][O[0]]);
    bkptr[0][i] = (int_4)-1;
  }

  // do the dynamic programming
  //
  for (int_4 i=1; i<T; i++) {
    for (int_4 j=0; j<S; j++) {

      // find the least cost bkptr
      //
      float_4 max_sum = s_err[i-1][0] + log10_cc(a[0][j]);
      int_4 index = 0;
      
      for (int_4 k=1; k<S; k++) {
	float_4 sum = s_err[i-1][k] + log10_cc(a[k][j]);
	if (sum > max_sum) {
	  max_sum = sum;
	  index = k;
	}
      }

      // update the cost and bkptr
      //
      s_err[i][j] = s_err[i-1][index] +
	log10_cc(a[index][j]) + log10_cc(b[j][O[i]]);
      bkptr[i][j] = index;
    }
  }

  // find the optimal endpoint
  //
  float_4 max_sum = s_err[T-1][0];
  int_4 index = 0;
      
  for (int_4 i=1; i<S; i++) {
    if (s_err[T-1][i] > max_sum) {
      max_sum = s_err[T-1][i];
      index = i;
    }
  }

  // provide debug information
  //
  if (debug > ISIP_DEBUG_BRIEF) {
    fprintf(stdout, "\n   dynamic programming information (time/state):\n");
    for (int_4 i=0; i<T; i++) {
      fprintf(stdout, "   O[%5ld] (%1c): ", i, y[O[i]]);
      for (int_4 j=0; j<S; j++)
	{fprintf(stdout, "%7.3f ", s_err[i][j]);}
      fprintf(stdout, "\n               : ");
      for (int_4 j=0; j<S; j++)
	{fprintf(stdout, "%7ld ", bkptr[i][j]);}
      fprintf(stdout, "\n\n");
    }

    if (debug > ISIP_DEBUG_DETAILED) {
      fprintf(stdout, "\tthe optimal path ends at state %ld\n", index);
    }
    fprintf(stdout, "   end of dynamic programming information.\n\n");
  }

  // backtrack to find the optimal state sequence
  //
  int_4 ptr = index;
  x[T-1] = ptr;

  for (int_4 i=T-1; i>0; i--) {
    ptr = bkptr[i][ptr];
    x[i-1] = ptr;
  }

  // exit gracefully
  //
  return max_sum;
}

//-----------------------------------------------------------------------------
// method: test_baumwelch_cc
//
// arguments:
//  int_4* x: (output) optimal state sequence
//  int_4* O: (input) observation sequence
//  int_4 T: (input) length of the input sequence
//  int_4 debug: (input) debug flag
//
// return: a float_4 value containing the log prob of the input sequence
//
//-----------------------------------------------------------------------------
float_4 Hmm::test_baumwelch_cc(int_4* x, int_4* O, int_4 T, int_4 debug) {

  // declare local variables
  //
  float_4 alpha[T][S];
  float_4 beta[T][S];
  float_4 c[T];

  // iterate over all data for alpha
  //
  for (int_4 t=0; t<T; t++) {
    for (int_4 j=0; j<S; j++) {

      // compute the inner sum
      //
      float_4 sum = 0.0;

      if (t == (int_4)0)
	{sum = pi[j];}
      else {
	for (int_4 i=0; i<S; i++)
	  {sum += alpha[t-1][i] * a[j][i];}
      }
    
      // update alpha
      //
      alpha[t][j] = sum * b[j][O[t]];
    }

    // compute the scaling constant
    //
    c[t] = 0;
    for (int_4 j=0; j<S; j++)
      {c[t] += alpha[t][j];}
    c[t] = invert_cc(c[t]);

    // scale the alpha
    //
    for (int_4 j=0; j<S; j++)
      {alpha[t][j] *= c[t];}
  }

  // iterate over all data for beta
  //
  for (int_4 t=T-1; t>=0; t--) {
    for (int_4 i=0; i<S; i++) {

      // compute the inner sum
      //
      float_4 sum = 0.0;

      if (t == (int_4)T-1)
	{sum = 1.0;}
      else {
	for (int_4 j=0; j<S; j++)
	  {sum += beta[t+1][j] * a[j][i] * b[j][O[t+1]];}
      }
    
      // update beta
      //
      beta[t][i] = sum * c[t];
    }
  }

  // compute P(y|M) using the scaling constant
  //
  float_4 sum_pi_den = 1.0;
  for (int_4 t=0; t<T; t++)
    {sum_pi_den *= c[t];}
  sum_pi_den = invert_cc(sum_pi_den);

  // provide debug information
  //
  if (debug > ISIP_DEBUG_BRIEF) {
    fprintf(stdout, "\n   baum welch information (time/state):\n");
    for (int_4 i=0; i<T; i++) {
      fprintf(stdout, "   O[%5ld] (%1c): ", i, y[O[i]]);
      for (int_4 j=0; j<S; j++)
	{fprintf(stdout, "%7.3f ", alpha[i][j]);}
      fprintf(stdout, "\n               : ");
      for (int_4 j=0; j<S; j++)
	{fprintf(stdout, "%7.3f ", beta[i][j]);}
      fprintf(stdout, "\n\n");
    }
    fprintf(stdout, "   end of baum welch information.\n\n");
  }

  // 'recognize' the sequence just for kicks to grab a state sequence
  // note: optimal state sequences don't have meaning for the baum welch
  // algorithm.
  //
  test_viterbi_cc(x, O, T, ISIP_DEBUG_NONE);

  // exit gracefully
  //
  return log10_cc(sum_pi_den);
}
