// file: hmm_train.cc
//

// isip include files
//
#include "../include/hmm.h"

#include "../include/integral_constants.h"
#include "../include/hmm_constants.h"

//-----------------------------------------------------------------------------
// method: train_symbol_table_cc
//
// arguments:
//  FILE* fp: (input) open file pointer
//
// return: a logical_1 value indicating status
//
// this method traverses a file and accumulates different looking symbols
// until K number of symbols are found. linefeeds are ignored.
// if more than K symbols are found, and error is generated.
//
//-----------------------------------------------------------------------------
logical_1 Hmm::train_symbol_table_cc(FILE* fp) {

  // make sure there is space
  //
  if (y == (char_1*)NULL)
    {return ISIP_FALSE;}

  // iterate over all data in the file
  //
  int_4 chr;
  int_4 num_sym_found = (int_4)0;
  rewind(fp);

  while ((chr = fgetc(fp)) != EOF) {
    if (chr != ISIP_NEWLINE) {

      // check if it is already on the list
      //
      logical_1 flag_match = ISIP_FALSE;
      for (int_4 i=0; i<num_sym_found; i++) {
	if (y[i] == chr) {
	  flag_match = ISIP_TRUE;
	  break;
	}
      }

      // add it to the list if there is no match
      //
      if (flag_match == ISIP_FALSE) {

	// make sure we haven't exceeded the expected
	// number of symbols
	//
	if (num_sym_found == K)
	  {return ISIP_FALSE;}
	else {
	  y[num_sym_found] = chr;
	  num_sym_found++;
	}
      }
    }
  }

  // rewind the file again
  //
  rewind(fp);

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: train_cc
//
// arguments:
//  int_4* x: (output) optimal state sequence
//  int_4* O: (input) observation sequence
//  int_4 T: (input) length of the input sequence
//  int_4 alg: (input) algorithm flag
//  int_4 debug: (input) debug flag
//
// return: a float_4 value containing the log prob of the input sequence
//
//-----------------------------------------------------------------------------
float_4 Hmm::train_cc(int_4* x, int_4* O, int_4 T, int_4 alg, int_4 debug) {

  // declare parameters
  //
  float_4 prob;

  // viterbi algorithm
  //
  if (alg == ISIP_ALG_VITERBI) {

    // 'recognize' the sequence
    //
    prob = test_viterbi_cc(x, O, T, debug);

    // update the counters
    //
    train_viterbi_cc(x, O, T, debug);
  }

  // baum-welch algorithm
  //
  else if (alg == ISIP_ALG_BAUMWELCH) {

    // evaluate the sequence
    //
    prob = train_baumwelch_cc(x, O, T, debug);
  }

  // error: unknown algorithm
  //
  else {
    fprintf(stdout, "<train_cc> unkown algorithm requested\n");
    prob = 1.0;
  }

  // exit gracefully
  //
  return prob;
}

//-----------------------------------------------------------------------------
// method: train_viterbi_cc
//
// arguments:
//  int_4* x: (input) optimal state sequence
//  int_4* O: (input) observation sequence
//  int_4 T: (input) length of the input sequence
//  int_4 debug: (input) debug flag
//
// return: a logical_1 value indicating status
//
// this method updates the counters used to estimate the new model
//
//-----------------------------------------------------------------------------
logical_1 Hmm::train_viterbi_cc(int_4* x, int_4* O, int_4 T, int_4) {

  // follow the optimal state sequence and update transition probs
  //
  for (int_4 i=0; i<T-1; i++)
    {a_new[x[i]][x[i+1]] += (float_4)1.0;}

  // follow the optimal state sequence and update observation probs
  //
  for (int_4 i=0; i<T; i++)
    {b_new[x[i]][O[i]] += (float_4)1.0;}

  // update the initial state prob
  //
  pi_new[x[0]] += (float_4)1.0;

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: replace_cc
//
// arguments:
//  int_4 alg: (input) algorithm flag
//
// return: a logical_1 value indicating status
//
//-----------------------------------------------------------------------------
logical_1 Hmm::replace_cc(int_4 alg) {

  // viterbi algorithm
  //
  if (alg == ISIP_ALG_VITERBI) {
    return replace_viterbi_cc();
  }

  // baum-welch algorithm
  //
  else if (alg == ISIP_ALG_BAUMWELCH) {
    return replace_baumwelch_cc();
  }

  // exit ungracefully
  //
  return ISIP_FALSE;
}

//-----------------------------------------------------------------------------
// method: replace_viterbi_cc
//
// arguments: none
//
// return: a logical_1 value indicating status
//
// this method replaces an old model with a new model using a method
// consistent with the viterbi algorithm. a primitive smoothing is implemented
// to prevent zero probabilities from propagating.
//
//-----------------------------------------------------------------------------
logical_1 Hmm::replace_viterbi_cc() {

  // follow the optimal state sequence and update transition probs
  //
  for (int_4 i=0; i<S; i++) {

    // count the number of events
    //
    float_4 sum = 0;
    for (int_4 j=0; j<S; j++)
      {sum += a_new[i][j];}
    if (sum == (float_4)0.0)
      {sum = 1.0;}

    // replace
    //
    for (int_4 j=0; j<S; j++) {
      a[i][j] = a_new[i][j] / sum;
      if (a[i][j] < (float_4)ISIP_MIN_PROB)
	{a[i][j] = ISIP_MIN_PROB;}
    }

    // renormalize
    //
    sum = 0;
    for (int_4 j=0; j<S; j++)
      {sum += a[i][j];}
    for (int_4 j=0; j<S; j++)
      {a[i][j] /= sum;}
  }

  // follow the optimal state sequence and update output probs
  //
  for (int_4 i=0; i<S; i++) {

    // count the number of events
    //
    float_4 sum = 0;
    for (int_4 j=0; j<K; j++)
      {sum += b_new[i][j];}
    if (sum == (float_4)0.0)
      {sum = 1.0;}

    // replace with clipping
    //
    for (int_4 j=0; j<K; j++) {
      b[i][j] = b_new[i][j] / sum;
      if (b[i][j] < (float_4)ISIP_MIN_PROB)
	{b[i][j] = ISIP_MIN_PROB;}
    }

    // renormalize
    //
    sum = 0;
    for (int_4 j=0; j<K; j++)
      {sum += b[i][j];}
    for (int_4 j=0; j<K; j++)
      {b[i][j] /= sum;}
  }

  // update the initial state probs
  //

  // count the number of events
  //
  float_4 sum = 0;
  for (int_4 i=0; i<S; i++)
    {sum += pi_new[i];}
  if (sum == (float_4)0.0)
    {sum = 1.0;}

  // replace
  //
  for (int_4 i=0; i<S; i++) {
    pi[i] = pi_new[i] / sum;
    if (pi[i] < (float_4)ISIP_MIN_PROB)
      {pi[i] = ISIP_MIN_PROB;}
  }

  // renormalize
  //
  sum = 0;
  for (int_4 i=0; i<S; i++)
    {sum += pi[i];}
  for (int_4 i=0; i<S; i++)
    {pi[i] /= sum;}

  // exit gracefully
  //
  return ISIP_TRUE;
}

//-----------------------------------------------------------------------------
// method: train_baumwelch_cc
//
// arguments:
//  int_4* x: (input) optimal state sequence
//  int_4* O: (input) observation sequence
//  int_4 T: (input) length of the input sequence
//  int_4 debug: (input) debug flag
//
// return: a float_4 value indicating the log prob of the data
//
// this method updates the counters used to estimate the new model
//
//-----------------------------------------------------------------------------
float_4 Hmm::train_baumwelch_cc(int_4* x, int_4* O, int_4 T, int_4 debug) {

  // declare local variables
  //
  float_4 alpha[T][S];
  float_4 beta[T][S];
  float_4 c[T];

  // iterate over all data for alpha
  //
  for (int_4 t=0; t<T; t++) {
    for (int_4 j=0; j<S; j++) {

      // compute the inner sum
      //
      float_4 sum = 0.0;

      if (t == (int_4)0)
	{sum = pi[j];}
      else {
	for (int_4 i=0; i<S; i++)
	  {sum += alpha[t-1][i] * a[j][i];}
      }
    
      // update alpha
      //
      alpha[t][j] = sum * b[j][O[t]];
    }

    // compute the scaling constant
    //
    c[t] = 0;
    for (int_4 j=0; j<S; j++)
      {c[t] += alpha[t][j];}
    c[t] = invert_cc(c[t]);

    // scale the alpha
    //
    for (int_4 j=0; j<S; j++)
      {alpha[t][j] *= c[t];}
  }

  // iterate over all data for beta
  //
  for (int_4 t=T-1; t>=0; t--) {
    for (int_4 i=0; i<S; i++) {

      // compute the inner sum
      //
      float_4 sum = 0.0;

      if (t == (int_4)T-1)
	{sum = 1.0;}
      else {
	for (int_4 j=0; j<S; j++)
	  {sum += beta[t+1][j] * a[j][i] * b[j][O[t+1]];}
      }
    
      // update beta
      //
      beta[t][i] = sum * c[t];
    }
  }

  // compute P(y|M) using the scaling constant
  //
  float_4 sum_pi_den = 1.0;
  for (int_4 t=0; t<T; t++)
    {sum_pi_den *= c[t];}
  sum_pi_den = invert_cc(sum_pi_den);

  // provide debug information
  //
  if (debug > ISIP_DEBUG_BRIEF) {
    fprintf(stdout, "\n   baum welch information (time/state):\n");
    for (int_4 i=0; i<T; i++) {
      fprintf(stdout, "   O[%5ld] (%1c): ", i, y[O[i]]);
      for (int_4 j=0; j<S; j++)
	{fprintf(stdout, "%7.3f ", alpha[i][j]);}
      fprintf(stdout, "\n               : ");
      for (int_4 j=0; j<S; j++)
	{fprintf(stdout, "%7.3f ", beta[i][j]);}
      fprintf(stdout, "\n\n");
    }
    fprintf(stdout, "   end of baum welch information.\n\n");
  }

  // update the transition prob counters
  //
  for (int_4 i=0; i<S; i++) {
    for (int_4 j=0; j<S; j++) {
      
      // accumulate the new contribution
      //
      float_4 sum_a = 0.0;
      //      float_4 sum_b = 0.0;
      for (int_4 t=0; t<T-2; t++)
	{sum_a += alpha[t][i] * a[j][i] * b[j][O[t+1]] * beta[t+1][j];}
      a_new[j][i] += sum_a / sum_pi_den;
    }
  }

  // update the output symbol prob counters
  //
  for (int_4 j=0; j<S; j++) {
    for (int_4 k=0; k<K; k++) {

      // accumulate the new contribution
      //
      float_4 sum_b = 0.0;
      for (int_4 t=0; t<T; t++) {
	if (O[t] == k)
	  {sum_b += alpha[t][j] * beta[t][j];}
      }
      b_new[j][k] += sum_b / sum_pi_den;
    }
  }

  // update the initial prob counters
  //
  for (int_4 i=0; i<S; i++)
    {pi_new[i] += (alpha[0][i] * beta[0][i]) / sum_pi_den;}

  // 'recognize' the sequence just for kicks to grab a state sequence
  // note: optimal state sequences don't have meaning for the baum welch
  // algorithm.
  //
  test_viterbi_cc(x, O, T, ISIP_DEBUG_NONE);

  // exit gracefully
  //
  return log10_cc(sum_pi_den);
}

//-----------------------------------------------------------------------------
// method: replace_baumwelch_cc
//
// arguments: none
//
// return: a logical_1 value indicating status
//
// this method replaces an old model with a new model using a method
// consistent with the baum-welch algorithm.
//
//-----------------------------------------------------------------------------
logical_1 Hmm::replace_baumwelch_cc() {

  // replace transitions
  //
  for (int_4 i=0; i<S; i++) {

    // compute the total number of transitions emanating from this node
    //
    float_4 sum_i = 0.0;
    for (int_4 j=0; j<S; j++)
      {sum_i += a_new[i][j];}
    if (sum_i == (float_4)0.0)
      {sum_i = 1.0;}

    // replace each transition
    //
    for (int_4 j=0; j<S; j++)
      {a[i][j] = a_new[i][j] / sum_i;}
  }

  // replace observations
  //
  for (int_4 j=0; j<S; j++) {

    // compute the total number of transitions emanating from this node
    //
    float_4 sum_j = 0.0;
    for (int_4 k=0; k<K; k++)
      {sum_j += b_new[j][k];}
    if (sum_j == (float_4)0.0)
      {sum_j = 1.0;}

    // replace each transition
    //
    for (int_4 k=0; k<K; k++)
      {b[j][k] = b_new[j][k] / sum_j;}
  }

  // update the initial state probs
  //

  // compute the total number of transitions
  //
  float_4 sum_i = 0.0;
  for (int_4 i=0; i<S; i++)
    {sum_i += pi_new[i];}

  for (int_4 i=0; i<S; i++)
    {pi[i] = pi_new[i] / sum_i;}

  // exit gracefully
  //
  return ISIP_TRUE;
}
